import numpy as np
import re
from Elasticipy.StressStrainTensors import SecondOrderTensor, StrainTensor, StressTensor
from Elasticipy.SphericalFunction import SphericalFunction, HyperSphericalFunction
from scipy.spatial.transform import Rotation


def _parse_tensor_components(prefix, **kwargs):
    pattern = r'^{}(\d{{2}})$'.format(prefix)
    value = dict()
    for k, v in kwargs.items():
        match = re.match(pattern, k)  # Extract 'C11' to '11' and so
        if match:
            value[match.group(1)] = v
    return value


def voigt(i, j):
    voigt_mat = np.array([[0, 5, 4],
                          [5, 1, 3],
                          [4, 3, 2]])
    return voigt_mat[i, j]


def unvoigt(i):
    inverse_voigt_mat = np.array([[0, 0],
                                  [1, 1],
                                  [2, 2],
                                  [1, 2],
                                  [0, 2],
                                  [0, 1]])
    return inverse_voigt_mat[i]


def _compute_unit_strain_along_direction(S, m, n, transverse=False):
    m_vec = np.atleast_2d(m)
    n_vec = np.atleast_2d(n)
    if not isinstance(S, ComplianceTensor):
        S = S.inv()
    if np.any(np.linalg.norm(m_vec) < 1e-9) or np.any(np.linalg.norm(n_vec) < 1e-9):
        raise ValueError('The input vector cannot be zeros')
    m_vec = (m_vec.T / np.linalg.norm(m_vec, axis=1)).T
    n_vec = (n_vec.T / np.linalg.norm(n_vec, axis=1)).T

    indices = np.indices((3, 3, 3, 3))
    i, j, k, ell = indices[0], indices[1], indices[2], indices[3]
    dot = np.abs(np.einsum('ij,ij->i', m_vec, n_vec))
    if np.any(np.logical_and(dot > 1e-9, dot < (1 - 1e-9))):
        raise ValueError('The two directions must be either equal or orthogonal.')
    if transverse:
        cosine = m_vec[:, i] * m_vec[:, j] * n_vec[:, k] * n_vec[:, ell]
    else:
        cosine = m_vec[:, i] * n_vec[:, j] * m_vec[:, k] * n_vec[:, ell]
    return np.einsum('pijkl,ijkl->p', cosine, S.full_tensor())


def tensorFromCrystalSymmetry(symmetry='Triclinic', point_group=None, diad='x',
                              tensor='Stiffness', phase_name='', **kwargs):
    """
    Create a fourth-order tensor from limited number of components, taking advantage of crystallographic symmetries

    Parameters
    ----------
    symmetry : str, default Triclinic
        Name of the crystallographic symmetry
    point_group : str
        Point group of the considered crystal. Only used (and mandatory) for tetragonal and trigonal symmetries.
    diad : str {'x', 'y'}, default 'x'
        Alignment convention. Sets whether x||a or y||b. Only used for monoclinic symmetry.
    tensor : str {'Stiffness', 'Compliance'}, default 'Stiffness'
        Type of tensor to create
    phase_name : str, default None
        Name to use when printing the tensor
    kwargs
        Keywords describing all the necessary components, depending on the crystal's symmetry and the type of tensor.
        For Stiffness, they should be named as 'Cij' (e.g. C11=..., C12=...).
        For Comliance, they should be named as 'Sij' (e.g. S11=..., S12=...).
        See examples below.

    Returns
    -------
    StiffnessTensor or ComplianceTensor

    Examples
    --------
    >>> tensorFromCrystalSymmetry(symmetry='monoclinic', diad='y', phase_name='TiNi',
    ...                          C11=231, C12=127, C13=104,
    ...                          C22=240, C23=131, C33=175,
    ...                          C44=81, C55=11, C66=85,
    ...                          C15=-18, C25=1, C35=-3, C46=3)
    Stiffness tensor (in Voigt notation) for TiNi:
    [[231. 127. 104.   0. -18.   0.]
     [127. 240. 131.   0.   1.   0.]
     [104. 131. 175.   0.  -3.   0.]
     [  0.   0.   0.  81.   0.   3.]
     [-18.   1.  -3.   0.  11.   0.]
     [  0.   0.   0.   3.   0.  85.]]
    Symmetry: monoclinic

    >>> tensorFromCrystalSymmetry(symmetry='monoclinic', diad='y', phase_name='TiNi',
    ...                          C11=8, C12=-3, C13=-2,
    ...                          C22=8, C23=-5, C33=10,
    ...                          C44=12, C55=116, C66=12,
    ...                          C15=14, C25=-8, C35=0, C46=0)
    Stiffness tensor (in Voigt notation) for TiNi:
    [[  8.  -3.  -2.   0.  14.   0.]
     [ -3.   8.  -5.   0.  -8.   0.]
     [ -2.  -5.  10.   0.   0.   0.]
     [  0.   0.   0.  12.   0.   0.]
     [ 14.  -8.   0.   0. 116.   0.]
     [  0.   0.   0.   0.   0.  12.]]
    Symmetry: monoclinic
    """
    tensor = tensor.lower()
    if tensor == 'stiffness':
        prefix = 'C'
        k = 1 / 2
    else:
        prefix = 'S'
        k = 2
    values = _parse_tensor_components(prefix, **kwargs)
    C = np.zeros((6, 6))
    symmetry = symmetry.lower()
    if ((symmetry == 'tetragonal') or (symmetry == 'trigonal')) and (point_group is None):
        raise ValueError('For tetragonal and trigonal symmetries, the point group is mandatory.')
    tetra_1 = ['-42m', '422', '4mm', '4/mm']
    tetra_2 = ['4', '-4', '4m']
    trigo_1 = ['32', '3m', '-3m']
    trigo_2 = ['3', '-3']
    if point_group is not None:
        if (point_group in tetra_1) or (point_group in tetra_2):
            symmetry = 'tetragonal'
        elif (point_group in trigo_1) or (point_group in trigo_1):
            symmetry = 'trigonal'
    try:
        if symmetry == 'isotropic':
            C[0, 0] = C[1, 1] = C[2, 2] = values['11']
            C[0, 1] = C[0, 2] = C[1, 2] = values['12']
            C[3, 3] = C[4, 4] = C[5, 5] = (C[0, 0] - C[0, 1]) * k
        elif symmetry == 'cubic':
            C[0, 0] = C[1, 1] = C[2, 2] = values['11']
            C[0, 1] = C[0, 2] = C[1, 2] = values['12']
            C[3, 3] = C[4, 4] = C[5, 5] = values['44']
        elif symmetry == 'hexagonal':
            C[0, 0] = C[1, 1] = values['11']
            C[0, 1] = values['12']
            C[0, 2] = C[1, 2] = values['13']
            C[2, 2] = values['33']
            C[3, 3] = C[4, 4] = values['44']
            C[5, 5] = (C[0, 0] - C[0, 1]) * k
        elif symmetry == 'tetragonal':
            C[0, 0] = C[1, 1] = values['11']
            C[0, 1] = values['12']
            C[0, 2] = C[1, 2] = values['13']
            C[2, 2] = values['33']
            C[3, 3] = C[4, 4] = values['44']
            C[5, 5] = values['66']
            if point_group in tetra_2:
                C[0, 5] = values['16']
                C[1, 5] = -C[0, 5]
        elif symmetry == 'trigonal':
            C[0, 0] = C[1, 1] = values['11']
            C[0, 1] = values['12']
            C[0, 2] = C[1, 2] = values['13']
            C[0, 3] = C[4, 5] = values['14']
            C[1, 3] = -C[0, 3]
            C[2, 2] = values['33']
            C[3, 3] = C[4, 4] = values['44']
            C[5, 5] = (C[0, 0] - C[0, 1]) * k
            if point_group in trigo_2:
                C[1, 4] = values['25']
                C[3, 5] = C[1, 4]
                C[0, 4] = -C[1, 4]
        else:  # Orthorombic, monoclinic or triclinic
            C[0, 0], C[0, 1], C[0, 2] = values['11'], values['12'], values['13']
            C[1, 1], C[1, 2] = values['22'], values['23']
            C[2, 2] = values['33']
            C[3, 3], C[4, 4], C[5, 5] = values['44'], values['55'], values['66']
            if (symmetry == 'monoclinic') or (symmetry == 'triclinic'):
                if (symmetry == 'monoclinic') and (diad == 'x'):
                    C[0, 5], C[1, 5], C[2, 5] = values['16'], values['26'], values['36']
                    C[3, 4] = values['45']
                else:
                    C[0, 4], C[1, 4], C[2, 4] = values['15'], values['25'], values['35']
                    C[3, 5] = values['46']
                if symmetry == 'triclinic':
                    C[0, 3], C[1, 3], C[2, 3] = values['14'], values['24'], values['34']
                    C[0, 4], C[1, 4], C[2, 4] = values['15'], values['25'], values['35']
                    C[3, 5], C[4, 5] = values['46'], values['56']
        if tensor == 'stiffness':
            constructor = StiffnessTensor
        else:
            constructor = ComplianceTensor
        return constructor(C + np.tril(C.T, -1), symmetry=symmetry, phase_name=phase_name)
    except KeyError as key:
        entry_error = prefix + key.args[0]
        if (symmetry == 'tetragonal') or (symmetry == 'trigonal'):
            err_msg = "For point group {}, keyword argument {} is required".format(point_group, entry_error)
        elif symmetry == 'monoclinic':
            err_msg = "For {} symmetry with diag='{}', keyword argument {} is required".format(symmetry, diad, entry_error)
        else:
            err_msg = "For {} symmetry, keyword argument {} is required".format(symmetry, entry_error)
        raise ValueError(err_msg)


class SymmetricTensor:
    """
    Template class for manipulating symmetric fourth-order tensors.

    Attributes
    ----------
    matrix : np.ndarray
        (6,6) matrix gathering all the components of the tensor, using the Voigt notation.
    symmetry : str
        Symmetry of the tensor

    """
    tensor_name = 'Symmetric'
    voigt_map = np.ones((6, 6))

    def __init__(self, M, phase_name='', symmetry='Triclinic', orientations=None):
        """
        Construct of stiffness tensor from a (6,6) matrix

        Parameters
        ----------
        M : np.ndarray
            (6,6) matrix corresponding to the stiffness tensor, written using the Voigt notation
        phase_name : str, default None
            Name to display
        symmetry : str, default Triclinic
        """
        self.matrix = np.array(M)
        self.phase_name = phase_name
        self.symmetry = symmetry
        self.orientations = orientations

    def __repr__(self):
        if self.phase_name == '':
            heading = '{} tensor (in Voigt notation):\n'.format(self.tensor_name)
        else:
            heading = '{} tensor (in Voigt notation) for {}:\n'.format(self.tensor_name, self.phase_name)
        print_symmetry = '\nSymmetry: {}'.format(self.symmetry)
        msg = heading + self.matrix.__str__() + print_symmetry
        if self.orientations is not None:
            msg = msg + '\n{} orientations'.format(len(self))
        return msg

    def __len__(self):
        if self.orientations is None:
            return ()
        elif self.orientations.single:
            return 1
        else:
            return len(self.orientations)

    def full_tensor(self):
        """
        Returns the full (unvoigted) tensor, as a [3, 3, 3, 3] array

        Returns
        -------
        np.ndarray
            Full tensor (4-index notation)
        """
        i, j, k, ell = np.indices((3, 3, 3, 3))
        ij = voigt(i, j)
        kl = voigt(k, ell)
        m = self.matrix[ij, kl] / self.voigt_map[ij, kl]
        if self.orientations is None:
            return m
        else:
            ori = self.orientations.as_matrix()
            rotated_tensors = np.einsum('qim,qjn,qko,qlp,mnop->qijkl', ori, ori, ori, ori, m)
            return rotated_tensors

    def rotate(self, rotation):
        """
        Apply a single rotation to a tensor, and return its component into the rotated frame.

        Parameters
        ----------
        rotation : Rotation
            Rotation to apply

        Returns
        -------
        SymmetricTensor
            Rotated tensor
        """
        rot_mat = rotation.as_matrix()
        rotated_tensor = np.einsum('im,jn,ko,lp,mnop->ijkl', rot_mat, rot_mat, rot_mat, rot_mat, self.full_tensor())
        ij, kl = np.indices((6, 6))
        i, j = unvoigt(ij).T
        k, ell = unvoigt(kl).T
        rotated_matrix = rotated_tensor[i, j, k, ell] * self.voigt_map[ij, kl]
        return self.__class__(rotated_matrix)

    def __add__(self, other):
        if isinstance(other, np.ndarray):
            if other.shape == (6, 6):
                mat = self.matrix + other
            elif other.shape == (3, 3, 3, 3):
                ten = self.full_tensor() + other
                ij, kl = np.indices((6, 6))
                i, j = unvoigt(ij).T
                k, ell = unvoigt(kl).T
                mat = ten[i, j, k, ell] * self.voigt_map[ij, kl]
            else:
                raise ValueError('The input argument must be either a 6x6 matrix or a (3,3,3,3) array.')
        elif isinstance(other, SymmetricTensor):
            if type(other) == type(self):
                mat = self.matrix + other.matrix
            else:
                raise ValueError('The two tensors to add must be of the same class.')
        else:
            raise ValueError('I don''t know how to add {} with {}.'.format(type(self), type(other)))
        return self.__class__(mat)

    def __mul__(self, other):
        if isinstance(other, SecondOrderTensor):
            return SecondOrderTensor(self * other.matrix)
        elif isinstance(other, np.ndarray):
            if other.shape[-2:] == (3, 3):
                if self.orientations is None:
                    return np.einsum('ijkl,...kl->...ij', self.full_tensor(), other)
                else:
                    return np.einsum('qijkl,...kl->q...ij', self.full_tensor(), other)
        elif isinstance(other, Rotation):
            if other.single:
                return self.rotate(other)
            else:
                return self.__class__(self.matrix, symmetry=self.symmetry, orientations=other)
        else:
            return self.__class__(self.matrix * other, symmetry=self.symmetry)

    def __rmul__(self, other):
        if isinstance(other, (Rotation, float, int)):
            return self * other
        else:
            raise ValueError('A fourth order tensor can be left-multiplied by rotations or scalar only.')

    def _orientation_average(self, orientations):
        """
        Rotate the tensor by a series of rotations, then evaluate its mean value.

        Parameters
        ----------
        orientations : np.ndarray or Rotation
            If an array is provided, it must be of shape [m, 3, 3], where orientations[i,:,:] gives i-th the orientation
            matrix

        Returns
        -------
        SymmetricTensor
            Mean tensor

        """
        if isinstance(orientations, Rotation):
            orientations = orientations.as_matrix()
        if len(orientations.shape) == 2:
            raise ValueError('The orientation must be a 3x3 or a Nx3x3 matrix')
        elif len(orientations.shape) == 2:
            return self * orientations
        else:
            m = orientations
            rotated_tensor = np.einsum('qim,qjn,qko,qlp,mnop->ijkl', m, m, m, m, self.full_tensor())
            ij, kl = np.indices((6, 6))
            i, j = unvoigt(ij).T
            k, ell = unvoigt(kl).T
            rotated_matrix = rotated_tensor[i, j, k, ell] * self.voigt_map[ij, kl] / orientations.shape[0]
            return self.__class__(rotated_matrix)


class StiffnessTensor(SymmetricTensor):
    """
    Class for manipulating fourth-order stiffness tensors.
    """
    tensor_name = 'Stiffness'

    def __init__(self, S, **kwargs):
        super().__init__(S, **kwargs)

    def __mul__(self, other):
        if isinstance(other, StrainTensor):
            new_tensor = super().__mul__(other)
            return StressTensor(new_tensor.matrix)
        elif isinstance(other, StressTensor):
            raise ValueError('You cannot multiply a stiffness tensor with a Stress tensor.')
        else:
            return super().__mul__(other)

    def inv(self):
        """
        Compute the reciprocal compliance tensor

        Returns
        -------
        ComplianceTensor
            Reciprocal tensor
        """
        C = np.linalg.inv(self.matrix)
        return ComplianceTensor(C, symmetry=self.symmetry, phase_name=self.phase_name, orientations=self.orientations)

    @property
    def Young_modulus(self):
        """
        Directional Young's modulus

        Returns
        -------
        SphericalFunction
            Young's modulus
        """
        def compute_young_modulus(n):
            eps = _compute_unit_strain_along_direction(self, n, n)
            return 1 / eps
        return SphericalFunction(compute_young_modulus)

    @property
    def shear_modulus(self):
        """
        Directional shear modulus

        Returns
        -------
        HyperSphericalFunction
            Shear modulus
        """
        def compute_shear_modulus(m, n):
            eps = _compute_unit_strain_along_direction(self, m, n)
            return 1 / (4 * eps)

        return HyperSphericalFunction(compute_shear_modulus)

    @property
    def Poisson_ratio(self):
        """
        Directional Poisson's ratio

        Returns
        -------
        HyperSphericalFunction
            Poisson's ratio
        """
        def compute_PoissonRatio(m, n):
            eps1 = _compute_unit_strain_along_direction(self, m, m)
            eps2 = _compute_unit_strain_along_direction(self, m, n, transverse=True)
            return -eps2 / eps1
        return HyperSphericalFunction(compute_PoissonRatio)

    def Voigt_average(self, orientations=None):
        """
        Compute the Voigt average of stiffness tensor

        Parameters
        ----------
        orientations : np.ndarray or None
            Set of m orientation matrices, defined as a [m, 3, 3] array.
            If None, uniform distribution is assumed, resulting in isotropic tensor

        Returns
        -------
        StiffnessTensor
            Voigt average of stiffness tensor

        See Also
        --------
        Reuss_average : compute the Reuss average
        Hill_average : compute the Voigt-Reuss-Hill average
        """
        if orientations is None:
            c = self.matrix
            C11 = (c[0, 0] + c[1, 1] + c[2, 2]) / 5 \
                + (c[0, 1] + c[0, 2] + c[1, 2]) * 2 / 15 \
                + (c[3, 3] + c[4, 4] + c[5, 5]) * 4 / 15
            C12 = (c[0, 0] + c[1, 1] + c[2, 2]) / 15 \
                + (c[0, 1] + c[0, 2] + c[1, 2]) * 4 / 15 \
                - (c[3, 3] + c[4, 4] + c[5, 5]) * 2 / 15
            C44 = (c[0, 0] + c[1, 1] + c[2, 2] - c[0, 1] - c[0, 2] - c[1, 2]) / 15 + (c[3, 3] + c[4, 4] + c[5, 5]) / 5
            mat = np.array([[C11, C12, C12, 0,   0,   0],
                            [C12, C11, C12, 0,   0,   0],
                            [C12, C12, C11, 0,   0,   0],
                            [0,   0,   0,   C44, 0,   0],
                            [0,   0,   0,   0,   C44, 0],
                            [0,   0,   0,   0,   0,   C44]])
            return StiffnessTensor(mat, symmetry='isotropic', phase_name=self.phase_name)
        else:
            return self._orientation_average(orientations)

    def Reuss_average(self, **kwargs):
        """
        Compute the Reuss average of tensor

        Parameters
        ----------
        orientations : np.ndarray or None
            Set of m orientation matrices, defined as a [m, 3, 3] array.
            If None, uniform distribution is assumed, resulting in isotropic tensor

        Returns
        -------
        StiffnessTensor
            Reuss average of stiffness tensor

        See Also
        --------
        Voigt_average : compute the Voigt average
        Hill_average : compute the Voigt-Reuss-Hill average
        """
        return self.inv().Reuss_average(**kwargs).inv()

    def Hill_average(self, orientations=None):
        """
        Compute the (Voigt-Reuss-)Hill average of tensor

        Parameters
        ----------
        orientations : np.ndarray or None
            Set of m orientation matrices, defined as a [m, 3, 3] array.
            If None, uniform distribution is assumed, resulting in isotropic tensor

        Returns
        -------
        StiffnessTensor
            Voigt-Reuss-Hill average of tensor

        See Also
        --------
        Voigt_average : compute the Voigt average
        Reuss_average : compute the Reuss average
        """
        Reuss = self.Reuss_average(orientations=orientations)
        Voigt = self.Voigt_average(orientations=orientations)
        return (Reuss + Voigt) * 0.5


class ComplianceTensor(StiffnessTensor):
    """
    Class for manipulating compliance tensors

    """
    tensor_name = 'Compliance'
    voigt_map = np.vstack((
        np.hstack((np.ones((3, 3)), 2 * np.ones((3, 3)))),
        np.hstack((2 * np.ones((3, 3)), 4 * np.ones((3, 3))))
    ))

    def __init__(self, C, **kwargs):
        super().__init__(C, **kwargs)

    def __mul__(self, other):
        if isinstance(other, StressTensor):
            return StrainTensor(self * other.matrix)
        elif isinstance(other, StrainTensor):
            raise ValueError('You cannot multiply a compliance tensor with Strain tensor.')
        else:
            return super().__mul__(other)

    def inv(self):
        """
        Compute the reciprocal stiffness tensor

        Returns
        -------
        StiffnessTensor
            Reciprocal tensor
        """
        S = np.linalg.inv(self.matrix)
        return StiffnessTensor(S, symmetry=self.symmetry, phase_name=self.phase_name, orientations=self.orientations)

    def Reuss_average(self, orientations=None):
        if orientations is None:
            s = self.matrix
            C11 = (s[0, 0] + s[1, 1] + s[2, 2]) / 5 \
                + (s[0, 1] + s[0, 2] + s[1, 2]) * 2 / 15 \
                + (s[3, 3] + s[4, 4] + s[5, 5]) * 1 / 15
            C12 = (s[0, 0] + s[1, 1] + s[2, 2]) / 15 \
                + (s[0, 1] + s[0, 2] + s[1, 2]) * 4 / 15 \
                - (s[3, 3] + s[4, 4] + s[5, 5]) * 1 / 30
            C44 = (s[0, 0] + s[1, 1] + s[2, 2] - s[0, 1] - s[0, 2] - s[1, 2]) * 4 / 15 + (s[3, 3] + s[4, 4] + s[5, 5]) / 5
            mat = np.array([[C11, C12, C12, 0,   0,   0],
                            [C12, C11, C12, 0,   0,   0],
                            [C12, C12, C11, 0,   0,   0],
                            [0,   0,   0,   C44, 0,   0],
                            [0,   0,   0,   0,   C44, 0],
                            [0,   0,   0,   0,   0,   C44]])
            return ComplianceTensor(mat, symmetry='isotropic', phase_name=self.phase_name)
        else:
            return self._orientation_average(orientations)

    def Voigt_average(self, **kwargs):
        return self.inv().Voigt_average(**kwargs).inv()

    def Hill_average(self, **kwargs):
        return self.inv().Hill_average(**kwargs)




import numpy as np
from Elasticipy.SecondOrderTensor import SecondOrderTensor


class StrainTensor(SecondOrderTensor):
    """
    Class for manipulating symmetric strain tensors or arrays of symmetric strain tensors.

    """
    name = 'Strain tensor'
    voigt_map = [1, 1, 1, 2, 2, 2]

    def principalStrains(self):
        """
        Values of the principals strains.

        If the tensor array is of shape [m,n,...], the results will be of shape [m,n,...,3].

        Returns
        -------
        np.ndarray
            Principal strain values
        """
        return self.eig()[0]

    def volumetricStrain(self):
        """
        Volumetric change (1st invariant of the strain tensor)

        Returns
        -------
        np.array or float
            Volumetric change
        """
        return self.I1


class StressTensor(SecondOrderTensor):
    """
    Class for manipulating stress tensors or arrays of stress tensors.
    """
    name = 'Stress tensor'

    def principalStresses(self):
        """
        Values of the principals stresses.

        If the tensor array is of shape [m,n,...], the results will be of shape [m,n,...,3].

        Returns
        -------
        np.ndarray
            Principal stresses
        """
        return np.real(self.eig()[0])

    @property
    def J1(self):
        """
        First invariant of the deviatoric part of the stress tensor. It is always zeros, as the deviatoric part as null
        trace.

        Returns
        -------
        float or np.ndarray
            zero(s)
        """
        if self.shape:
            return np.zeros(self.shape)
        else:
            return 0.0

    @property
    def J2(self):
        """
        Second invariant of the deviatoric part of the stress tensor.

        Returns
        -------
        float or np.ndarray
            J2 invariant
        """
        return -self.deviatoricPart().I2

    @property
    def J3(self):
        """
        Third invariant of the deviatoric part of the stress tensor.

        Returns
        -------
        float or np.ndarray
            J3 invariant
        """
        return self.deviatoricPart().I3

    def vonMises(self):
        """
        von Mises equivalent stress.

        Returns
        -------
        np.ndarray or float
            von Mises equivalent stress

        See Also
        --------
        Tresca : Tresca equivalent stress
        """
        return np.sqrt(3 * self.J2)

    def Tresca(self):
        """
         Tresca(-Guest) equivalent stress.

         Returns
         -------
         np.ndarray or float
             Tresca equivalent stress

        See Also
        --------
        vonMises : von Mises equivalent stress
        """
        ps = self.principalStresses()
        return np.max(ps, axis=-1) - np.min(ps, axis=-1)

    def hydrostaticPressure(self):
        """
        Hydrostatic pressure

        Returns
        -------
        np.ndarray or float

        See Also
        --------
        sphericalPart : spherical part of the stress
        """
        return -self.I1/3



import numpy as np
from scipy.spatial.transform import Rotation


class _MatrixProxy:
    def __init__(self, matrix):
        self.matrix = matrix

    def __getitem__(self, args):
        sub = self.matrix[(...,) + (args if isinstance(args, tuple) else (args,))]
        if sub.shape == ():
            return float(sub)
        else:
            return sub

    def __setitem__(self, args, value):
        self.matrix[(...,) + (args if isinstance(args, tuple) else (args,))] = value


class SecondOrderTensor:
    """
    Template class for manipulation of second order tensors or arrays of second order tensors

    Attributes
    ----------
    matrix : np.ndarray
        (...,3,3) matrix storing all the components of the tensor

    """
    name = 'Second-order tensor'
    "Name to use when printing the tensor"

    voigt_map = [1, 1, 1, 1, 1, 1]
    "List of factors to use for building a tensor from Voigt vector(s)"

    def __init__(self, matrix):
        """
        Create an array of second-order tensors.

        The input argument can be:
            - an array of shape (3,3) defining all the component of the tensor;
            - a stack of matrices, that is an array of shape (...,3,3);
            - an array of shape (6);
            - a stack of vectors of lenghts 6, that is an array of shape (...,6).
        In the two last cases, it is assumed that the Voigt numbering convention is used.

        Parameters
        ----------
        matrix : list or np.ndarray
            (3,3) matrix, stack of (3,3) matrices, 6-length vector or stack of 6-length vectors
        """
        matrix = np.array(matrix)
        shape = matrix.shape
        if shape and (shape[-1] == 6):
            new_shape = shape[:-1] + (3, 3)
            unvoigted_matrix = np.zeros(new_shape)
            voigt = [[0, 0], [1, 1], [2, 2], [1, 2], [0, 2], [0, 1]]
            for i in range(6):
                unvoigted_matrix[..., voigt[i][0], voigt[i][1]] = matrix[..., i]/self.voigt_map[i]
                unvoigted_matrix[..., voigt[i][1], voigt[i][0]] = matrix[..., i]/self.voigt_map[i]
            self.matrix = unvoigted_matrix
        elif len(shape) > 1 and shape[-2:] == (3, 3):
            self.matrix = matrix
        else:
            raise ValueError('The input matrix must be of shape (3,3) or (...,3,3)')

    def __repr__(self):
        s = self.name + '\n'
        if self.shape:
            s += 'Shape={}'.format(self.shape)
        else:
            s += self.matrix.__str__()
        return s

    def __getitem__(self, index):
        return self.__class__(self.matrix[index])

    def __setitem__(self, index, value):
        if isinstance(value, (float, np.ndarray)):
            self.matrix[index] = value
        elif type(value) == self.__class__:
            self.matrix[index] = value.matrix
        else:
            raise ValueError('The r.h.s must be either float, a ndarray or an object of class {}'.format(self.__class__))

    def __add__(self, other):
        if type(self) == type(other):
            return self.__class__(self.matrix + other.matrix)
        elif isinstance(other, (int, float, np.ndarray)):
            return self.__class__(self.matrix + other)
        else:
            raise ValueError('The element to add must be a number, a ndarray or of the same class.')

    def __sub__(self, other):
        if type(self) == type(other):
            return self.__class__(self.matrix - other.matrix)
        elif isinstance(other, (int, float, np.ndarray)):
            return self.__class__(self.matrix - other)
        else:
            raise ValueError('The element to subtract must be a number, a ndarray or of the same class.')

    @property
    def shape(self):
        """
        Return the shape of the tensor array

        Returns
        -------
        tuple
            Shape of array

        See Also
        --------
        ndim : number of dimensions
        """
        *shape, _, _ = self.matrix.shape
        return tuple(shape)

    @property
    def ndim(self):
        """
        Return the number of dimensions of the tensor array

        Returns
        -------
        float
            number of dimensions

        See Also
        --------
        shape : shape of tensor array
        """
        return len(self.shape)

    @property
    def C(self):
        """
        Return tensor components

        For instance T.C[i,j] returns all the (i,j)-th components of each tensor in the array.

        Returns
        -------
        np.ndarray
            Tensor components
        """
        return _MatrixProxy(self.matrix)

    def eig(self):
        """
        Eigenvalues of the tensor

        Returns
        -------
        lambda : np.ndarray
            Eigenvalues of each tensor.
        v : np.ndarray
            Eigenvectors of teach tensor.

        See Also
        --------
        principalDirections : return only the principal directions (without eigenvalues)
        """
        return np.linalg.eig(self.matrix)

    def principalDirections(self):
        """
        Principal directions of the tensors

        Returns
        -------
        np.nparray
            Principal directions of each tensor of the tensor array

        See Also
        --------
        eig : Return both eigenvalues and corresponding principal directions
        """
        return self.eig()[1]

    @property
    def I1(self):
        """
        First invariant of the tensor (trace)

        Returns
        -------
        np.ndarray or float
            First invariant(s) of the tensor(s)

        See Also
        --------
        I2 : Second invariant of the tensors
        I3 : Third invariant of the tensors (det)
        """
        return self.matrix.trace(axis1=-1, axis2=-2)

    @property
    def I2(self):
        """
        Second invariant of the tensor

        For a matrix M, it is defined as::

            I_2 = 0.5 * ( np.trace(M)**2 + np.trace(np.matmul(M, M.T)) )

        Returns
        -------
        np.array or float
            Second invariant(s) of the tensor(s)

        See Also
        --------
        I1 : First invariant of the tensors (trace)
        I3 : Third invariant of the tensors (det)
        """
        a = self.I1**2
        b = np.matmul(self.matrix, self._transposeTensor()).trace(axis1=-1, axis2=-2)
        return 0.5 * (a - b)

    @property
    def I3(self):
        """
        Third invariant of the tensor (determinant)

        Returns
        -------
        np.array or float
            Third invariant(s) of the tensor(s)

        See Also
        --------
        I1 : First invariant of the tensors (trace)
        I2 : Second invariant of the tensors
        """
        return np.linalg.det(self.matrix)

    def trace(self):
        """
        Return the traces of the tensor array

        Returns
        -------
        np.ndarray or float
            traces of each tensor of the tensor array

        See Also
        --------
        I1 : First invariant of the tensors (trace)
        I2 : Second invariant of the tensors
        I3 : Third invariant of the tensors (det)
        """
        return self.I1

    def __mul__(self, B):
        """
        Element-wise matrix multiplication of arrays of tensors. Each tensor of the resulting tensor array is computed
        as the matrix product of the tensor components.

        Parameters
        ----------
        B : SecondOrderTensor or np.ndarray or Rotation or float
            If B is a numpy array, we must have::

                B.shape == (..., 3, 3)

        Returns
        -------
            Array of tensors populated with element-wise matrix multiplication.

        See Also
        --------
        matmul : matrix-like multiplication of tensor arrays
        """
        if isinstance(B, SecondOrderTensor):
            return SecondOrderTensor(np.matmul(self.matrix, B.matrix))
        elif isinstance(B, Rotation):
            rotation_matrices = B.as_matrix()
            transpose_matrices = np.swapaxes(rotation_matrices, -1, -2)
            new_matrix = np.matmul(np.matmul(transpose_matrices, self.matrix), rotation_matrices)
            return self.__class__(new_matrix)
        elif isinstance(B, (float, int)):
            return self.__class__(self.matrix * B)
        elif isinstance(B, np.ndarray):
            if B.shape == self.shape:
                new_matrix = np.einsum('...ij,...->...ij', self.matrix, B)
                return self.__class__(new_matrix)
            elif B.shape == self.matrix.shape:
                return self.__class__(np.matmul(self.matrix, B))
            else:
                err_msg = 'For a tensor of shape {}, the input argument must be an array of shape {} or {}'.format(
                    self.shape, self.shape, self.shape + (3, 3))
                raise ValueError(err_msg)
        else:
            raise ValueError('The input argument must be a tensor, an ndarray, a rotation or a scalar value.')

    def matmul(self, other):
        """
        Perform matrix-like product between tensor arrays. Each "product" is a matrix product between
        the tensor components.

        If A.shape=(a1, a2, ..., an) and B.shape=(b1, b2, ..., bn), with C=A.matmul(B), we have::

            C.shape = (a1, a2, ..., an, b1, b2, ..., bn)

        and::

            C[i,j,k,...,p,q,r...] = np.matmul(A[i,j,k,...], B[p,q,r,...])

        Parameters
        ----------
        other : SecondOrderTensor or np.ndarray or Rotation
            Tensor array or rotation to right-multiply by. If Rotation is provided, the rotations are applied on each
            tensor.

        Returns
        -------
        SecondOrderTensor
            Tensor array

        See Also
        --------
        __mul__ : Element-wise matrix product
        """
        if isinstance(other, SecondOrderTensor):
            other_matrix = other.matrix
        elif isinstance(other, Rotation):
            other_matrix = other.as_matrix()
        else:
            other_matrix = other
        matrix = self.matrix
        shape_matrix = matrix.shape[:-2]
        shape_other = other_matrix.shape[:-2]
        extra_dim_matrix = len(shape_other)
        extra_dim_other = len(shape_matrix)
        matrix_expanded = matrix.reshape(shape_matrix + (1,) * extra_dim_other + (3, 3))
        other_expanded = other_matrix.reshape((1,) * extra_dim_matrix + shape_other + (3, 3))
        if isinstance(other, Rotation):
            other_expanded_t = np.swapaxes(other_expanded, -1, -2)
            new_mat = np.matmul(np.matmul(other_expanded_t, matrix_expanded), other_expanded)
        else:
            new_mat = np.matmul(matrix_expanded, other_expanded)
        return self.__class__(np.squeeze(new_mat))

    def transposeArray(self):
        """
        Transpose the array of tensors

        If A is a tensor array of shape [s1, s2, ..., sn], A.T is of shape [sn, ..., s2, s1].

        Returns
        -------
        SecondOrderTensor
            Transposed array

        See Also
        --------
        T : transpose the tensor array (not the components)
        """
        if self.ndim < 2:
            return self
        else:
            matrix = self.matrix
            ndim = matrix.ndim
            new_axes = np.hstack((ndim - 3 - np.arange(ndim - 2), -2, -1))
            transposed_arr = np.transpose(matrix, new_axes)
            return self.__class__(transposed_arr)

    @property
    def T(self):
        """
        Transpose the array of tensors.

        It is actually an alias for transposeArray()

        Returns
        -------
        SecondOrderTensor
            Transposed array
        """
        return self.transposeArray()

    def _transposeTensor(self):
        return np.swapaxes(self.matrix, -1, -2)

    def transposeTensor(self):
        """
        Transpose of tensors of the tensor array

        Returns
        -------
        SecondOrderTensor
            Array of transposed tensors of the tensor array

        See Also
        --------
        Transpose : transpose the array (not the components)
        """
        return self.__class__(self._transposeTensor())

    def ddot(self, other):
        """
        Double dot product (contraction of tensor product, usually denoted ":") of two tensors.

        For two tensors whose matrices are M1 and M2::

            M1.ddot(M2) == np.trace(np.matmul(M1, M2))

        Parameters
        ----------
        other : SecondOrderTensor or np.ndarray
            Tensor or tensor array to multiply by before contraction.

        Returns
        -------
        float or np.ndarray
            Result of double dot product

        See Also
        --------
        matmul : matrix-like product between two tensor arrays.

        """
        tensor_prod = self*other
        return tensor_prod.trace()

    def _flatten(self):
        if self.shape:
            new_len = np.prod(self.shape)
            return np.reshape(self.matrix, (new_len, 3, 3))
        else:
            return self.matrix

    def _stats(self, fun, axis=None):
        if axis is None:
            flat_mat = self._flatten()
            new_matrix = fun(flat_mat, axis=0)
        else:
            if axis < 0:
                axis += -2
            if (axis > self.ndim - 1) or (axis < -self.ndim - 2):
                raise ValueError('The axis index is out of bounds for tensor array of shape {}'.format(self.shape))
            new_matrix = fun(self.matrix, axis=axis)
        return self.__class__(new_matrix)

    def flatten(self):
        """
        Flatten the array of tensors.

        If T is of shape [s1, s2, ..., sn], the shape for T.flatten() is [s1*s2*...*sn].

        Returns
        -------
        SecondOrderTensor
            Flattened array (vector) of tensor

        See Also
        --------
        ndim : number of dimensions of the tensor array
        shape : shape of the tensor array
        """
        return self.__class__(self._flatten())

    def mean(self, axis=None):
        """
        Arithmetic mean value

        Parameters
        ----------
        axis : int or None, default None
            Axis to compute the mean along with.
            If None, returns the overall mean (mean of flattened array)

        Returns
        -------
        SecondOrderTensor
            Mean tensor

        See Also
        --------
        std : Standard deviation
        min : Minimum value
        max : Maximum value
        """
        if self.ndim:
            return self._stats(np.mean, axis=axis)
        else:
            return self

    def std(self, axis=None):
        """
          Standard deviation

          Parameters
          ----------
          axis : int or None, default None
              Axis to compute standard deviation along with.
              If None, returns the overall standard deviation (std of flattened array)

          Returns
          -------
          SecondOrderTensor
              Tensor of standard deviation

        See Also
        --------
        mean : Mean value
        min : Minimum value
        max : Maximum value
          """
        if self.ndim:
            return self._stats(np.std, axis=axis)
        else:
            return self.__class__(np.zeros((3, 3)))

    def min(self, axis=None):
        """
        Minimum value

        Parameters
        ----------
        axis : int or None, default None
           Axis to compute minimum along with.
           If None, returns the overall minimum (min of flattened array)

        Returns
        -------
        SecondOrderTensor
           Minimum value of tensors

        See Also
        --------
        max : Maximum value
        mean : Mean value
        std : Standard deviation
        """
        if self.ndim:
            return self._stats(np.min, axis=axis)
        else:
            return self

    def max(self, axis=None):
        """
        Maximum value

        Parameters
        ----------
        axis : int or None, default None
            Axis to compute maximum along with.
            If None, returns the overall maximum (max of flattened array)

        Returns
        -------
        SecondOrderTensor
            Maximum value of tensors

        See Also
        --------
        min : Minimum value
        mean : Mean value
        std : Standard deviation
        """
        if self.ndim:
            return self._stats(np.max, axis=axis)
        else:
            return self

    def symmetricPart(self):
        """
        Symmetric part of the tensor

        Returns
        -------
        SecondOrderTensor
            Symmetric tensor

        See Also
        --------
        skewPart : Skew-symmetric part of the tensor
        """
        new_mat = 0.5 * (self.matrix + self._transposeTensor())
        return self.__class__(new_mat)

    def skewPart(self):
        """
        Skew-symmetric part of the tensor

        Returns
        -------
        SecondOrderTensor
            Skew-symmetric tensor
        """
        new_mat = 0.5 * (self.matrix - self._transposeTensor())
        return self.__class__(new_mat)

    def sphericalPart(self):
        """
        Spherical (hydrostatic) part of the tensor

        Returns
        -------
        SecondOrderTensor
            Spherical part

        See Also
        --------
        I1 : compute the first invariant of the tensor
        deviatoricPart : deviatoric the part of the tensor
        """
        s = self.I1 / 3
        return self.eye(self.shape)*s

    def deviatoricPart(self):
        """
        Deviatoric part of the tensor

        Returns
        -------
        SecondOrderTensor

        See Also
        --------
        sphericalPart : spherical part of the tensor
        """
        return self - self.sphericalPart()

    @classmethod
    def eye(cls, shape=()):
        """
        Create an array of tensors populated with identity matrices

        Parameters
        ----------
        shape : list or int, default ()
            If not provided, it just creates a single identity tensor. Otherwise, the tensor array will be of the
            specified shape.

        Returns
        -------
        SecondOrderTensor
            Array of identity tensors

        See Also
        --------
        ones : creates an array of tensors full of ones
        zeros : creates an array full of zero tensors
        """
        if isinstance(shape, int):
            matrix_shape = (shape, 3, 3)
        else:
            matrix_shape = shape + (3, 3,)
        eye = np.zeros(matrix_shape)
        eye[..., np.arange(3), np.arange(3)] = 1
        return cls(eye)

    @classmethod
    def ones(cls, shape=()):
        """
        Create an array of tensors populated with matrices of full of ones.

        Parameters
        ----------
        shape : list or int, default ()
            If not provided, it just creates a single tensor of ones. Otherwise, the tensor array will be of the
            specified shape.

        Returns
        -------
        SecondOrderTensor
            Array of ones tensors

        See Also
        --------
        eye : creates an array of identity tensors
        zeros : creates an array full of zero tensors
        """
        if isinstance(shape, int):
            matrix_shape = (shape, 3, 3)
        else:
            matrix_shape = shape + (3, 3,)
        ones = np.ones(matrix_shape)
        return cls(ones)

    @classmethod
    def zeros(cls, shape=()):
        """
        Create an array of tensors populated with matrices full of zeros.

        Parameters
        ----------
        shape : list or int, default ()
            If not provided, it just creates a single tensor of ones. Otherwise, the tensor array will be of the
            specified shape.

        Returns
        -------
        SecondOrderTensor
            Array of ones tensors

        See Also
        --------
        eye : creates an array of identity tensors
        ones : creates an array of tensors full of ones
        """
        if isinstance(shape, int):
            matrix_shape = (shape, 3, 3)
        else:
            matrix_shape = shape + (3, 3,)
        zeros = np.zeros(matrix_shape)
        return cls(zeros)



import unittest
from pytest import approx
import numpy as np
from scipy.spatial.transform import Rotation

from FourthOrderTensor import StiffnessTensor
import StressStrainTensors as Tensors


Cmat = [[231, 127, 104, 0, -18, 0],
        [127, 240, 131, 0, 1, 0],
        [104, 131, 175, 0, -3, 0],
        [0, 0, 0, 81, 0, 3],
        [-18, 1, -3, 0, 11, 0],
        [0, 0, 0, 3, 0, 85]]
C = StiffnessTensor(Cmat)


class TestStrainStressTensors(unittest.TestCase):
    def test_mult_by_stiffness(self):
        """
        Test Stiffness/Strain tensors product C*eps (which stands for C:E)
        """
        tensile_dir = [1, 0, 0]
        stress = Tensors.StressTensor([[1, 0, 0],
                                       [0, 0, 0],
                                       [0, 0, 0]])
        strain = C.inv()*stress
        eps_xx = strain.C[0, 0]
        eps_yy = strain.C[1, 1]
        eps_zz = strain.C[2, 2]
        E = C.Young_modulus.eval(tensile_dir)
        nu_y = C.Poisson_ratio.eval(tensile_dir, [0, 1, 0])
        nu_z = C.Poisson_ratio.eval(tensile_dir, [0, 0, 1])
        assert eps_xx == approx(1/E)
        assert eps_yy == approx(-nu_y / E)
        assert eps_zz == approx(-nu_z / E)

    def test_rotate_tensor(self, n_oris=10):
        """
        Test the rotation of a tensor

        Parameters
        ----------
        n_oris : int
            Number of random orientations to use
        """
        random_tensor = Tensors.SecondOrderTensor(np.random.random((3, 3)))
        random_oris = Rotation.random(n_oris)
        eps_rotated = random_tensor * random_oris
        for i in range(n_oris):
            rot_mat = random_oris.as_matrix()[i, :, :]
            eps_matrix_th = np.matmul(np.matmul(rot_mat.T, random_tensor.matrix), rot_mat)
            np.testing.assert_almost_equal(eps_matrix_th, eps_rotated[i].matrix)

    def test_transpose_array(self, shape=(1, 2, 3)):
        """
        Test transposing a tensor array

        Parameters
        ----------
        shape : tuple
            Shape of the tensor array to use
        """
        random_matrix = np.random.random(shape + (3, 3))
        random_tensor = Tensors.SecondOrderTensor(random_matrix)
        transposed_tensor = random_tensor.transposeArray()
        for i in range(shape[0]):
            for j in range(shape[1]):
                for k in range(shape[2]):
                    np.testing.assert_array_equal(random_matrix[i, j, k], transposed_tensor[k, j, i].matrix)

    def test_mul(self, shape=(4, 5)):
        """
        Test the element-wise product of tensors.

        Parameters
        ----------
        shape : tuple
            Shape of the tensor arrays to use.
        """
        shape = shape + (3, 3)
        matrix1 = np.random.random(shape)
        matrix2 = np.random.random(shape)
        tensor_prod = Tensors.SecondOrderTensor(matrix1) * Tensors.SecondOrderTensor(matrix2)
        for i in range(shape[0]):
            for j in range(shape[1]):
                mat_prod = np.matmul(matrix1[i, j], matrix2[i, j])
                np.testing.assert_array_equal(tensor_prod[i, j].matrix, mat_prod)

    def test_matmul(self, length1=3, length2=4):
        """
        Test the matrix-like product of tensor arrays

        Parameters
        ----------
        length1 : int
            Length of the first array
        length2 : int
            Length of the second array
        """
        matrix1 = np.random.random((length1, 3, 3))
        matrix2 = np.random.random((length2, 3, 3))
        rand_tensor1 = Tensors.SecondOrderTensor(matrix1)
        rand_tensor2 = Tensors.SecondOrderTensor(matrix2)
        cross_prod_tensor = rand_tensor1.matmul(rand_tensor2)
        for i in range(0, length1):
            for j in range(0, length2):
                mat_prod = np.matmul(matrix1[i], matrix2[j])
                np.testing.assert_array_equal(cross_prod_tensor[i, j].matrix, mat_prod)

    def test_statistics(self, shape=(5, 4, 3, 2)):
        """
        Test the std, min and max functions for tensor arrays.

        Parameters
        ----------
        shape : tuple
            Shape of the tensor to use
        """
        matrix = np.random.random(shape + (3, 3))
        tensor = Tensors.SecondOrderTensor(matrix)
        mini = tensor.min()
        maxi = tensor.max()
        std = tensor.std()
        # First, check T.std()
        for i in range(0, 3):
            for j in range(0, 3):
                Cij = matrix[..., i, j].flatten()
                assert np.std(Cij) == approx(std.C[i, j])
                assert np.min(Cij) == approx(mini.C[i, j])
                assert np.max(Cij) == approx(maxi.C[i, j])
        # Then, check T.std(axis=...)
        for i in range(0, len(shape)):
            np.testing.assert_array_equal(tensor.std(axis=i).matrix, np.std(matrix, axis=i))
            np.testing.assert_array_equal(tensor.min(axis=i).matrix, np.min(matrix, axis=i))
            np.testing.assert_array_equal(tensor.max(axis=i).matrix, np.max(matrix, axis=i))

    def test_ddot(self, shape=(4, 3, 2)):
        """
        Test the ddot method.

        Parameters
        ----------
        shape : tuple
            shape of the tensor arrays to use
        """
        matrix1 = np.random.random(shape + (3, 3))
        matrix2 = np.random.random(shape + (3, 3))
        tens1 = Tensors.SecondOrderTensor(matrix1)
        tens2 = Tensors.SecondOrderTensor(matrix2)
        ddot = tens1.ddot(tens2)
        for i in range(0, shape[0]):
            for j in range(0, shape[1]):
                for k in range(0, shape[2]):
                    ddot_th = np.trace(np.matmul(matrix1[i, j, k], matrix2[i, j, k]))
                    self.assertEqual(ddot_th, ddot[i, j, k])

    def test_vonMises_Tresca(self):
        """
        Check that the Tresca and von Mises methods work well for simple tension, simple shear and hydrostatic
        pressure.
        """
        matrix = np.zeros((3, 3, 3))
        matrix[0, 0, 0] = 1  # Simple tension
        matrix[1, 1, 0] = matrix[1, 0, 1] = 1  # Simple shear
        matrix[2, np.arange(3), np.arange(3)] = -1  # Hydrostatic pressure
        stress = Tensors.StressTensor(matrix)

        vM_stress = stress.vonMises()
        vm_th = np.array([1, np.sqrt(3), 0.0])
        np.testing.assert_array_equal(vM_stress, vm_th)

        Tresca_stress = stress.Tresca()
        Tresca_th = np.array([1, 2, 0.0])
        np.testing.assert_array_equal(Tresca_stress, Tresca_th)

    def test_rotation_stiffness(self, n_strain=50, n_ori=100):
        """
        Check that the two ways to compute stress from a rotated stiffness tensor are consistent.

        Parameters
        ----------
        n_strain : int
            Number of strain tensors to consider
        n_ori : int
            Number of random orientation to consider
        """
        matrix = np.random.random((n_strain, 3, 3))
        eps = Tensors.StrainTensor(matrix).symmetricPart()  # Ensure that the strain is symmetric
        ori = Rotation.random(n_ori)
        C_rotated = C * ori
        sigma = C_rotated * eps

        # Rotate stress and stress by their own
        eps_rot = eps.matmul(ori)
        sigma_rot2 = C * eps_rot
        sigma2 = sigma_rot2 * ori.inv()
        np.testing.assert_almost_equal(sigma.matrix, sigma2.matrix)

    def test_multidimensional_tensors(self, shape_strain=(5, 4, 3), n_ori=100):
        """
        Check that the shape of (C * rotations) * eps is (m, p, r, ...) if rotation.shape=(p, q,...) and
        len(rotations)=m.

        Parameters
        ----------
        shape_strain : tuple
            shape of tensor array
        n_ori : int
            number of orientations to consider
        """
        strain = Tensors.StrainTensor.ones(shape_strain)
        ori = Rotation.random(n_ori)
        C_rotated = C * ori
        stress = C_rotated * strain
        self.assertEqual(stress.shape, (n_ori,) + shape_strain)


if __name__ == '__main__':
    unittest.main()



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='Elasticipy',
    version='1.0.3',
    packages=find_packages(where="src"),
    package_dir={'': 'src'},
    url='https://github.com/DorianDepriester/Elasticipy',
    license='MIT Licence',
    author='Dorian Depriester',
    author_email='dorian.depriester@ensam.eu',
    description='Collection of tools to work on strain, stress and stiffness tensors, with plotting features',
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
            'scipy',
            'numpy',
        ],
)
